import {
   joinTypes, typeOfDependency, Dependency, BBinary, ImpType, ImpReturns
} from '../core/Dispatcher.js'

export type Complex<T> = {re: T; im: T;}

export const Complex_type = {
   test: <T>(dep: {testT: (z: unknown) => z is T}) =>
      (z: unknown): z is Complex<T> =>
      typeof z === 'object' && 're' in z && 'im' in z
      && dep.testT(z.re) && dep.testT(z.im),
   infer: (dep: typeOfDependency) =>
      (z: Complex<unknown>) => joinTypes(dep.typeOf(z.re), dep.typeOf(z.im)),
   from: {
      T: <T>(dep: Dependency<'zero', [T]>) => (t: T) =>
         ({re: t, im: dep.zero(t)}),
      Complex: <U,T>(dep: {convert: (from: U) => T}) =>
         (z: Complex<U>) => ({re: dep.convert(z.re), im: dep.convert(z.im)})
   }
}

type Binary<B> = [B, B]

export interface ComplexReturn<Params> {
   complex: Params extends [infer U] ? Complex<U> // unary case
      : Params extends BBinary<infer B> ? Complex<B> // binary case
      : never 
}

export const complex_unary =
   <T>(dep: Dependency<'zero', [T]>): ImpType<'complex', [T]> =>
   t => ({re: t, im: dep.zero(t)})
export const complex_binary = <T>(t: T, u: T): ImpReturns<'complex', [T,T]> =>
   ({re: t, im: u})
